#include "initX.h"

/*****************************************************************************
* Local typedefs
******************************************************************************/

/*
 * The number of color cells on each side of theColorCube.  This number
 * cubed will give the total number of cells.  We don't need too many,
 * since there are usually only 256 colors available, but it allows us
 * to handle all the colormapped and grayscale visuals in the same way,
 * and to reduce aliasing problems if we have a wierd number of colors.
 */ 

#define theCUBESIZE 16
#define theGRAYSIZE 1024

#define theWAITTIME  10  /* Wait up to this many seconds to write to screen */
#define theQUERYFREQ 10  /* How often to check the colormap for updates */

#define theEVENTMASK  (ButtonPressMask      | \
                       ButtonReleaseMask    | \
                       KeyPressMask         | \
		       KeyReleaseMask       | \
                       ExposureMask         | \
                       StructureNotifyMask  | \
		       VisibilityChangeMask | \
		       PointerMotionMask    )
/*		       ResizeRedirectMask)*/

#define theRED   0
#define theGREEN 1
#define theBLUE  2

#define TRUE 1
#define FALSE 0

/* Which command-line parameters for the X Window Display have been set */
#define theCLASSSET    0x0001
#define theWIDTH       0x0004                                                                    
#define theHEIGHT      0x0008                                                                    
#define theSTARTX      0x0010
#define theSTARTY      0x0020
#define theNEGATIVEX   0x0040
#define theNEGATIVEY   0x0080
#define theICONSTATE   0x0100
#define thePRIVATECMAP 0x0200
#define theRESIZABLE   0x0400

typedef struct mHandlerStruct {
  mWindow *W;
  void (*func)(mWindow *,XEvent *);
  struct mHandlerStruct *next;
} mHandler;

/*****************************************************************************
* Local variables
******************************************************************************/
 
/*
 * These are the settings for the actual display and window.
 */

Display      *theDisplay = NULL;    /* which display on this system */
int           theScreen;            /* which screen number of display */
int	      theDepth;
int           theClass=-1;
Visual 	     *theVisual=NULL; 	
int           theXVal=0;             /* x position to start at  */
int           theYVal=0;             /* y position to start at  */
unsigned int  theBorderWidth=4;      /* width of window borders */

static int	     theDispWidth, theDispHeight;
static Atom          WM_PROTOCOLS;         /* a window manager protocol msg */
static Atom          WM_DELETE_WINDOW;     /* if window manager kills app */
static Atom          THR_STOP_HANDLER;     
static int           theParameters=0;
/*
 * These are for the color quantization.
 */

static unsigned long theColorShift[3];      /* RGB offsets */
static unsigned long theColorSize[3];       /* RGB bits */
static Bool          fastbig;               /* if we have 8x8x8 RGB TrueColor */
static Bool          fastlittle;            /* if we have 8x8x8 BGR TrueColor */
static Colormap      theColormap = None;    /* our window's color map */
static unsigned int  theCells;              /* the number of colormap cells */
static int           theMaxCell = 1;        /* Maximum index we can allocate */
static unsigned long theMaxDist;            /* Maximum color error allowed */
static XColor       *theColorCube = NULL;   /* the color cube/gray ramp */
static XColor       *theColors = NULL;      /* the color map installed */

static mHandler     *theHandlers=NULL;
static int          handlerRunning=0;

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

#define _LOCK   pthread_mutex_lock(&mutex);
#define _UNLOCK pthread_mutex_unlock(&mutex);

static pthread_t       theHandlerID;

/*****************************************************************************
* Local Functions
******************************************************************************/
static void         _mWSelectVisual(Visual **theVisual, unsigned int *theDepth);
static void         _mWBuildColormap(Visual *theVisual);
static void         _mWGetBestColormap(Visual *theVisual);
static void         _mWGetBestGraymap(Visual *theVisual);
static int          _mWError(char *format,...);
static int          _mWMAError(char *msg,int val);
static void         _mWGetBestIndex(long r, long g, long b, XColor *theCell);
static XColor      *_mWGetBestColor (long r, long g, long b);
static void        *_mWHandleEvents(void *ptr);
/*****************************************************************************
* Static Functions
******************************************************************************/
/*****************************************************************************
* Function mCreatePThread
******************************************************************************/
int mCreatePThread(pthread_t *newID, void *(*func)(void *), void *arg) {
 pthread_attr_t attr;                                                                       
 int thr_stat;  

 pthread_attr_init(&attr);                                                                
 pthread_attr_setscope(&attr,PTHREAD_SCOPE_SYSTEM);                                       
 pthread_attr_setstacksize(&attr,THR_STACKSIZE); 
 thr_stat=pthread_create( newID, &attr, func, arg); 
 pthread_attr_destroy(&attr);  
 return thr_stat;                                                           
}

/*****************************************************************************
* Function mSetHandler
******************************************************************************/
void  mSetHandler(mWindow *W, void (*func)(mWindow *,XEvent *) ){
  mHandler *H;

  mStopHandler(W);
  if ((H=(mHandler *)malloc(sizeof(mHandler)))==NULL)
    _mWMAError("handler",sizeof(mHandler));
    
  H->W=W;
  H->func=func;
  H->next=theHandlers;  

  _LOCK
    theHandlers=H;
  _UNLOCK

  if (!handlerRunning) {
    handlerRunning=1;
    mCreatePThread(&theHandlerID,_mWHandleEvents,(void *)(theHandlers));
  }  

}

/*****************************************************************************
* Function mStopHandler
******************************************************************************/
void  mStopHandler(mWindow *W){
  mHandler *H,*P;
  
  if (!theHandlers) return;
  _LOCK
    if(theHandlers&&(theHandlers->W==W)) {
      H=theHandlers->next;
      free(theHandlers);
      theHandlers=H;
    }else if (theHandlers) 
      for (P=theHandlers,H=theHandlers->next;H;H=H->next,P=P->next) 
        if (H->W==W) {
	  P->next=H->next;
	  free(H);
	  break;
	}
  _UNLOCK  
}

/*****************************************************************************
* Function mInitX
******************************************************************************/
void mInitX(char *theDisplayName, int ownCmap, int handlers){
  XGCValues            theGCValues;

  /* we allow only one initialization */
  if (theDisplay) return;
  
  if (handlers) XInitThreads();
  if (ownCmap) theParameters|=thePRIVATECMAP;

  /* check for visual */
  if ( (theClass== StaticGray)||(theClass== GrayScale)||(theClass== StaticColor)||
       (theClass== PseudoColor) ||(theClass== TrueColor)||(theClass== DirectColor) ){
    theParameters |= theCLASSSET;
  }
  /* open display */
  if ((theDisplay = XOpenDisplay(theDisplayName)) == NULL) 
    _mWError("Error opening X display \"%s\".\n",(theDisplayName)?theDisplayName:"default");

  theScreen     = DefaultScreen(theDisplay);
  theDispWidth  = DisplayWidth(theDisplay, theScreen);
  theDispHeight = DisplayHeight(theDisplay, theScreen);

  _mWSelectVisual(&theVisual, &theDepth);
  _mWBuildColormap(theVisual);
}


/*****************************************************************************
* Function mEndX
******************************************************************************/
void mEndX() {
  if (theColormap != None)  {
    XFreeColormap(theDisplay, theColormap);
  }
  theMaxCell = 1;
  XCloseDisplay(theDisplay);
  theDisplay = NULL; 
  if (theColorCube != NULL) {
    free(theColorCube);
    theColorCube = NULL;
  }
  if (theColors != NULL)  {
    free(theColors);
    theColors = NULL;
  }
}

/*****************************************************************************
* Function mOpenWindow
******************************************************************************/
mWindow *mOpenWindow(int theWidth, int theHeight, int resizable,char *theTitle) {
  mWindow *W;
  XSetWindowAttributes theWindowAttributes;
  unsigned long        theWindowMask;
  XSizeHints           theSizeHints;
  XWMHints             theWMHints;
  XClassHint           theClassHints;
  XGCValues            theGCValues;                                                              
  int                  theParameters   =theHEIGHT|theWIDTH;
  

  if ((W=(mWindow *)malloc(sizeof(mWindow)))==NULL)
    _mWMAError("mWindow",sizeof(mWindow));
  
  /* check if a specific position is requested */
  if (theXVal!=0) {
    theParameters|=theSTARTX;
    if (theXVal<0) theParameters|=theNEGATIVEX;
  }
  if (theYVal!=0) {
    theParameters|=theSTARTY;
    if (theYVal<0) theParameters|=theNEGATIVEY;
  }
   
   /*
   * Accept only Expose events while creating the window.
   * It is reccommended to accept these events, in case we miss them
   * before we begin to accept all the other events.  Also ask for a
   * backing store so we don't have to refresh the window all the time.
   */

  theWindowAttributes.backing_store    = WhenMapped;
  theWindowAttributes.background_pixel = BlackPixel(theDisplay, theScreen);
  theWindowAttributes.border_pixel     = BlackPixel(theDisplay, theScreen);
  theWindowAttributes.colormap         = theColormap;
  theWindowAttributes.event_mask       = ExposureMask;
  theWindowMask = CWBackingStore | CWBackPixel | CWBorderPixel |
                  CWColormap | CWEventMask;

  /*
   * take the user starting geometry into account
   */

  if (theParameters & theSTARTX)
  {
    if (theParameters & theNEGATIVEX)
    {
      theXVal = theDispWidth - theWidth - theXVal - theBorderWidth;
      theParameters ^= theNEGATIVEX;
    }
  }
  else
  {
    theXVal = (theDispWidth - theWidth) / 2 - theBorderWidth;
  }

  if (theXVal < 0)   /* Make sure the title bar is on the screen */
    theXVal = 0;

  if (theParameters & theSTARTY)
  {
    if (theParameters & theNEGATIVEY)
    {
      theYVal = theDispHeight - theHeight - theYVal - theBorderWidth;
      theParameters ^= theNEGATIVEY;
    }
  }
  else
  {
    theYVal = (theDispHeight - theHeight) / 2 - theBorderWidth;
  }

  if (theYVal < 0)   /* Make sure the title bar is on the screen */
    theYVal = 0;


  W->theWindow = XCreateWindow(theDisplay, RootWindow(theDisplay, theScreen),
                            theXVal, theYVal,
                            (unsigned int)theWidth,
                            (unsigned int)theHeight,
                            theBorderWidth, (int)theDepth, InputOutput,
                            theVisual, theWindowMask, &theWindowAttributes);

  /*
   * tell the window manager what size of window we want
   */

  W->resizable=resizable;
  
  if (theParameters & theWIDTH || theParameters & theHEIGHT){
    theSizeHints.flags     |= USSize;
  }
  if (theParameters & theSTARTX || theParameters & theSTARTY)  {
    theSizeHints.x          = theXVal;            /* Obsolete */
    theSizeHints.y          = theYVal;            /* Obsolete */
    theSizeHints.flags     |= PPosition | USPosition;
  }

  theSizeHints.width        =                     /* Obsolete */
  theSizeHints.base_width   = theWidth;
  theSizeHints.height       =                     /* Obsolete */
  theSizeHints.base_height  = theHeight;
  
  theSizeHints.flags |= PBaseSize;
  if (!resizable) {
    theSizeHints.min_width    =
    theSizeHints.max_width    =
    theSizeHints.min_aspect.x =
    theSizeHints.max_aspect.x = theWidth;
  
    theSizeHints.min_height   =
    theSizeHints.max_height   =
    theSizeHints.min_aspect.y =
    theSizeHints.max_aspect.y = theHeight;
    theSizeHints.flags = PSize | PMinSize | PMaxSize | PAspect;
  } 
  else theSizeHints.flags = 0; /*PSize?*/
  XSetWMNormalHints(theDisplay, W->theWindow, &theSizeHints);


  /*
   * tell the window manager what title to use for this window
   */

  if (theTitle) XStoreName(theDisplay, W->theWindow, theTitle);

  /*
   * Now that we are finished setting everything up, we will begin
   * handling input for this window.
   */
    
  XSelectInput(theDisplay, W->theWindow, theEVENTMASK);

  /*
   * This tells the window manager that we want to know if the window is
   * being killed so we can catch it and do an orderly shutdown.
   */

  WM_DELETE_WINDOW = XInternAtom(theDisplay, "WM_DELETE_WINDOW", False);
  WM_PROTOCOLS = XInternAtom(theDisplay, "WM_PROTOCOLS", False);
  
  XSetWMProtocols(theDisplay, W->theWindow, &WM_DELETE_WINDOW, 1);

  /*
   * Now, could we please see the window on the screen?  Until now, we have
   * dealt with a window which has been created but has not appeared on the
   * screen.  Mapping the window places it visibly on the screen.
   */

  XMapWindow(theDisplay, W->theWindow);
  {XEvent e;
    while(1) {
      XWindowEvent(theDisplay,W->theWindow,theEVENTMASK,&e);
      if (e.type==MapNotify) break;
    }
    
  } 
  /* create a graphics context for drawing */
  theGCValues.function = GXcopy;
  W->theGC = XCreateGC(theDisplay, W->theWindow, GCFunction, &theGCValues);
  return W;
}


/*****************************************************************************
* Function mCloseWindow
******************************************************************************/
void mCloseWindow(mWindow *W) {
  if (!W) return;
  XFreeGC(theDisplay, W->theGC);
  XDestroyWindow(theDisplay, W->theWindow);
  XFlush(theDisplay);
}

/*****************************************************************************
* Function mSetSize
******************************************************************************/
void   mSetSize(mWindow *W,int width,int height){
  XSizeHints           theSizeHints;
  XRectangle R;
  
  theSizeHints.flags     |= USSize;
  theSizeHints.width        =                     /* Obsolete */
  theSizeHints.base_width   = width;
  theSizeHints.height       =                     /* Obsolete */
  theSizeHints.base_height  = height;
  theSizeHints.flags |= PBaseSize;

  if (!W->resizable) {
    theSizeHints.min_width    =
    theSizeHints.max_width    =
    theSizeHints.min_aspect.x =
    theSizeHints.max_aspect.x = width;
  
    theSizeHints.min_height   =
    theSizeHints.max_height   =
    theSizeHints.min_aspect.y =
    theSizeHints.max_aspect.y = height;
    theSizeHints.flags = PSize | PMinSize | PMaxSize | PAspect;
  } 
  else theSizeHints.flags = 0; /*PSize?*/
  XSetWMNormalHints(theDisplay, W->theWindow, &theSizeHints);

  XResizeWindow(theDisplay,W->theWindow,width,height);
  {XEvent e; XWindowEvent(theDisplay, W->theWindow, VisibilityChangeMask, &e);}  

  R.x=R.y=0;
  R.width=width;
  R.height=height;
  XSetClipRectangles(theDisplay, W->theGC, 0, 0, &R, 1, Unsorted);
  
}


/*****************************************************************************
* Function mSetBackgroundPixmap
******************************************************************************/
void mSetBackgroundPixmap(mWindow *W,Pixmap P) {
  XSetWindowBackgroundPixmap(theDisplay,W->theWindow,P); 
  XClearWindow(theDisplay,W->theWindow);
} 


/*****************************************************************************
* Function mCreateImage
******************************************************************************/
XImage *mCreateImage(int width,int height){
 XImage *theImage;
 long size;
 
 theImage = XCreateImage(theDisplay, theVisual, theDepth, ZPixmap, 0, NULL,
                          (unsigned int)width,
                          (unsigned int)height,
                          BitmapPad(theDisplay), 0);

  size = theImage->bytes_per_line * height;

  if ((theImage->data = (char *)calloc(1, size)) == NULL)
    _mWMAError("X image data", size);
  return theImage;
}

 
/*****************************************************************************
* Function mPixel
******************************************************************************/
unsigned long mPixel(mColor color) {
  unsigned char red, green, blue;
  unsigned long thePixel=0;
  XColor *theColor;
  long wantColor[3], gotColor[3];
  int  i;
   
  if (fastlittle) return (unsigned long)color;   

  red   = color>>16;
  green = (color & 0x00ff00)>>8;
  blue  = color & 0x0000ff;

  if (fastbig) return mRGB(green,blue,red);

  wantColor[theRED]   = (long)red   * 0x100 ;
  wantColor[theGREEN] = (long)green * 0x100 ;
  wantColor[theBLUE]  = (long)blue  * 0x100 ;

    /* clamp the colors to the valid range */

    for (i = 0; i < 3; i++)  {
      wantColor[i] = (wantColor[i] < 0) ? 0 :
                     ((wantColor[i] > 0xffff) ? 0xffff : wantColor[i]);
    }

    switch (theClass)
    {
      case StaticGray:
      case GrayScale:
      case StaticColor:
      case PseudoColor:

        /* get the best color match for the calculated pixel value */

        theColor = _mWGetBestColor(wantColor[theRED],
                                wantColor[theGREEN],
                                wantColor[theBLUE]);
        thePixel = theColor->pixel;

        gotColor[theRED]   = theColor->red;
        gotColor[theGREEN] = theColor->green;
        gotColor[theBLUE]  = theColor->blue;
        break;

      case TrueColor:
      case DirectColor:

        /*
         * First work out the color in the range needed for this visual.
         */

        gotColor[theRED]  = wantColor[theRED]  *theColorSize[theRED]  /0x10000;
        gotColor[theGREEN]= wantColor[theGREEN]*theColorSize[theGREEN]/0x10000;
        gotColor[theBLUE] = wantColor[theBLUE] *theColorSize[theBLUE] /0x10000;

        thePixel=((gotColor[theRED]   << theColorShift[theRED]  ) |
                  (gotColor[theGREEN] << theColorShift[theGREEN]) |
                  (gotColor[theBLUE]  << theColorShift[theBLUE] ));

        /*
         * Now, convert it back to a color in the range of [0-0xffff], so
         * we can add any truncation effects that were encountered
         * during the color conversion back into the error array.*/
         

        gotColor[theRED]  = gotColor[theRED]  * 0x10000/theColorSize[theRED];
        gotColor[theGREEN]= gotColor[theGREEN]* 0x10000/theColorSize[theGREEN];
        gotColor[theBLUE] = gotColor[theBLUE] * 0x10000/theColorSize[theBLUE];
        break;
    }
 /*   fprintf(stderr,"asking for color %u %u %u, got %u %u %u\n",
    wantColor[theRED],wantColor[theGREEN],wantColor[theBLUE],
    gotColor[theRED],gotColor[theGREEN],gotColor[theBLUE]);*/
    return thePixel;
  
} 



/*****************************************************************************
*
* FUNCTION  SelectVisual
*
* INPUT
*
* OUTPUT
*
*   global variable theClass is set
*
* RETURNS
*
*   theVisual - which visual we have chosen
*
* AUTHOR
*
* DESCRIPTION
*
*   Find the "best" visual, using this criteria:
*
*   1) Use the user specified visual class, if it is available.
*   2) Use the default visual if it has the deepest depth.
*   3) Deepest TrueColor visual
*   4) Deepest DirectColor visual
*   5) Deepest PseudoColor visual
*   6) Deepest StaticColor visual
*   8) Deepest StaticGray visual
*   9) Default visual
*
* CHANGES
*
*   Added tests for PseudoColor visuals other than 8 bits, StaticColor [AED]
*
******************************************************************************/

static void _mWSelectVisual(Visual **theVisual, unsigned int *theDepth){
  XVisualInfo *theDefaultVisualInfo, *theVisualInfo, *theBestVisualInfo, *p;
  XVisualInfo template;
  int theDeepestDepth = 0;
  int nItems;
  int i;

  /*
   * Obtain the default VisualInfo for the in case we don't find a better one.
   */
 
  *theVisual = DefaultVisual(theDisplay, theScreen);
  template.visualid = (*theVisual)->visualid;
  if ((theBestVisualInfo = theDefaultVisualInfo =
       XGetVisualInfo(theDisplay, VisualIDMask, &template, &nItems)) == NULL){
    _mWError("\nError obtaining X visual info for %s.\n",
           XDisplayString(theDisplay));
  }

  /*
   * Get a list of all visuals on the screen.
   */

  template.screen = theScreen;
  if ((theVisualInfo =
     XGetVisualInfo(theDisplay, VisualScreenMask, &template, &nItems)) == NULL){
    _mWError("\nError obtaining X visual info for %s.\n",
           XDisplayString(theDisplay));
  }

  *theDepth = 0;

  /*
   * Look for the visual of the user specified class or use the default
   * visual if it has the deepest depth.  We'll gain nothing but color
   * flashing if we use a different visual/colormap at the same depth.
   */

  if (theParameters & theCLASSSET) {
    for (i = 0, p = theVisualInfo; i < nItems; i++, p++)
      if ((p->class == theClass) && (p->depth > *theDepth)) {
        *theDepth = p->depth;
        theBestVisualInfo = p;
      }

/*    if (*theDepth == 0) {
      Status_Info("\nUnable to get ");
      switch (theClass) {
        case TrueColor:
          Status_Info("TrueColor");
          break;
        case DirectColor:
          Status_Info("DirectColor");
          break;
        case PseudoColor:
          Status_Info("PseudoColor");
          break;
        case StaticColor:
          Status_Info("StaticColor");
          break;
        case GrayScale:
          Status_Info("GrayScale");
          break;
        case StaticGray:
          Status_Info("StaticGray");
          break;
      }
      Status_Info(" visual as requested.");
    }*/
  }

  /*
   * If we can't get the requested visual, find the deepest available depth.
   * If this is the same as the default depth, then use the default visual.
   */

  if (*theDepth == 0) {
    for (i = 0; i < nItems; i++)
      if (theVisualInfo[i].depth > theDeepestDepth)
        theDeepestDepth = theVisualInfo[i].depth;
    if (theDeepestDepth == DefaultDepth(theDisplay, theScreen))
      *theDepth = theDeepestDepth;
  }

  /*
   * Look for deepest TrueColor
   */

  if (*theDepth == 0){
    for (i = 0, p = theVisualInfo; i < nItems; i++, p++){
      if ((p->class == TrueColor) && (p->depth > *theDepth)){
        *theDepth = p->depth;
        theBestVisualInfo = p;
      }
    }
  }

  /*
   * Look for deepest DirectColor
   */

  if (*theDepth == 0)
  {
    for (i = 0, p = theVisualInfo; i < nItems; i++, p++)
    {
      if ((p->class == DirectColor) && (p->depth > *theDepth))
      {
        *theDepth = p->depth;
        theBestVisualInfo = p;
      }
    }
  }

  /*
   * Look for deepest PseudoColor
   */

  if (*theDepth == 0)
  {
    for (i = 0, p = theVisualInfo; i < nItems; i++, p++)
    {
      if ((p->class == PseudoColor) && (p->depth > *theDepth))
      {
        *theDepth = p->depth;
        theBestVisualInfo = p;
      }
    }
  }

  /*
   * Look for deepest StaticColor
   */

  if (*theDepth == 0)
  {
    for (i = 0, p = theVisualInfo; i < nItems; i++, p++)
    {
      if ((p->class == StaticColor) && (p->depth > *theDepth))
      {
        *theDepth = p->depth;
        theBestVisualInfo = p;
      }
    }
  }

  /*
   * Look for deepest StaticGray
   */

  if (*theDepth == 0)
  {
    for (i = 0, p = theVisualInfo; i < nItems; i++, p++)
    {
      if ((p->class == StaticGray) && (p->depth > *theDepth))
      {
        *theDepth = p->depth;
        theBestVisualInfo = p;
      }
    }
  }

  *theDepth = theBestVisualInfo->depth;
  *theVisual = theBestVisualInfo->visual;
  if (*theVisual != DefaultVisual(theDisplay, theScreen))
    theParameters |= thePRIVATECMAP;
  theCells = theBestVisualInfo->colormap_size;
  theClass = theBestVisualInfo->class;
  theParameters |= theCLASSSET;  /* Save for next time if we are animating */

  /* Set up the RGB color masks for the given bit depth */
  if ((theClass == TrueColor) || (theClass == DirectColor))
  {
    unsigned long a;

    theColorShift[theRED] = 0;

    for (a = theBestVisualInfo->red_mask; (a & 1) == 0; a >>= 1)
    {
      theColorShift[theRED]++;
    }

    theColorSize[theRED]   = (theBestVisualInfo->red_mask >>
                                                theColorShift[theRED]) + 1;

    theColorShift[theGREEN] = 0;

    for (a = theBestVisualInfo->green_mask; (a & 1) == 0; a >>= 1)
      theColorShift[theGREEN]++;

    theColorSize[theGREEN] = (theBestVisualInfo->green_mask >>
                                                theColorShift[theGREEN]) + 1;

    theColorShift[theBLUE] = 0;

    for (a = theBestVisualInfo->blue_mask; (a & 1) == 0; a >>= 1)
      theColorShift[theBLUE]++;

    theColorSize[theBLUE]  = (theBestVisualInfo->blue_mask >>
                                                theColorShift[theBLUE]) + 1;

    /*
     * See if we have a normal 24 bit Truecolor visual, so we don't have
     * to do lots of math when displaying each pixel.
     */

    fastbig = True;
    fastlittle = True;

    for (a = 0; a < 3; a++) {
      if (theColorShift[a] != a*8 || theColorSize[a] != 0x100) fastbig = False;
      if (theColorShift[a] != (2 - a)*8 || theColorSize[a] != 0x100) fastlittle = False;
    }

    /*
     * We need to do this because if the visual has an alpha channel,
     * this is included in the depth, which we don't want to consider.
     */

    a = theColorSize[theRED] * theColorSize[theGREEN] *
                               theColorSize[theBLUE];
    theDeepestDepth = 0;

    while (a > 1)
    {
      a >>= 1;
      theDeepestDepth++;
    }
  }

  fprintf(stderr,"Using ");
  switch (theClass)
  {
    case TrueColor:
      fprintf(stderr,"%d bit TrueColor", theDeepestDepth);
      break;
    case DirectColor:
      fprintf(stderr,"%d bit DirectColor", theDeepestDepth);
      break;
    case PseudoColor:
      fprintf(stderr,"%d color PseudoColor", theCells);
      break;
    case StaticColor:
      fprintf(stderr,"%d color StaticColor", theCells);
      break;
    case GrayScale:
      fprintf(stderr,"%d level GrayScale", theCells);
      break;
    case StaticGray:
      fprintf(stderr,"%d level StaticGray", theCells);
      break;
  }
  fprintf(stderr," visual...\n");

  XFree((void *)theDefaultVisualInfo);
  XFree((void *)theVisualInfo);
}



/*****************************************************************************
*
* FUNCTION  BuildColormap
*
* INPUT
*
*   theVisual - type of visual we are using
*
* OUTPUT
*
*   global variables theColormap and theColorCube set and contents filled in
*
* RETURNS
*
* AUTHOR      Andreas Dilger   Feb, 1996
*
* DESCRIPTION
*
*   Create a Colormap structure.  To make life easier, we will always use a
*   color cube for PseudoColor and StaticColor, and a grayscale ramp for
*   Grayscale, and StaticGray visuals, as well as PseudoColor and StaticColor
*   visuals with the +dG option.  This color cube/gray ramp will be filled in
*   with the best approximation XColor values, so we will not have to search
*   the color space for each pixel to find the best colormap index, we can
*   still do a good job with Static visuals, and we can treat all of the
*   paletted visuals in the same way.  
*
* CHANGES
*
******************************************************************************/

static void _mWBuildColormap(Visual * theVisual){

  switch (theClass)
  {
    int rmask, gmask, bmask;
    int i, j;

    case PseudoColor:

      /*
       * Find the largest number of entries to fit in the given colormap.
       * This currently handles colormaps up to 8 bits (256 entries), with
       * a few spaces left over for the system colors.
       */

      theColorSize[theRED] = theColorSize[theGREEN] = theColorSize[theBLUE] = 6;

      for (i = theColorSize[theRED]; i > 2; i--)
      {
        for (j = 0; j < 3; j++)
        {
          if (theColorSize[theRED] * theColorSize[theGREEN] *
                                     theColorSize[theBLUE] <= theCells)
          {
            break;
          }

          theColorSize[j]--;
        }
      }

      /* This is the maximum number of new colormap entries we will allocate */
      theMaxCell = theColorSize[theRED] * theColorSize[theGREEN] *
                   theColorSize[theBLUE];
      _mWGetBestColormap(theVisual);
      break;

    case StaticColor:
      theColormap = XCreateColormap(theDisplay,
                                    RootWindow(theDisplay, theScreen),
                                    theVisual, AllocNone);

      theMaxCell = theCells;
      theMaxDist = 0x7fffffff;
      if ((theColors = (XColor *)calloc(theCells, sizeof(XColor))) == NULL)
        _mWMAError("X local colormap", theCells * sizeof(XColor));
      for (i = 0; i < theCells; i++)
      {
        theColors[i].pixel = i;
      }
      if ((theColorCube = (XColor *)calloc(theCUBESIZE*theCUBESIZE*theCUBESIZE,
                                           sizeof(XColor))) == NULL)
        _mWMAError("X color cube",
                theCUBESIZE*theCUBESIZE*theCUBESIZE*sizeof(XColor));
      break;

    case GrayScale:

      /*
       * Try to find the largest number of colors that will fit in the
       * colormap, with a few left over for other applications.  The largest
       * reasonable frame buffer is 10 bits, so we start with 1000 entries
       * and work our way down.   The color size should be at least 2 in
       * the end, for a total of 8 entries.  I don't think such a thing as
       * a 1- or 2-bit GrayScale visual exists (ie not StaticGray).
       */

      theColorSize[theRED]=theColorSize[theGREEN]=theColorSize[theBLUE] = 10;

      for (i = theColorSize[theRED]; i > 2; i--)
      {
        if (theColorSize[theRED] * theColorSize[theGREEN] *
                                   theColorSize[theBLUE] <= theCells)
        {
          break;
        }
        theColorSize[theRED]--;
        theColorSize[theGREEN]--;
        theColorSize[theBLUE]--;
      }

      /* This is the maximum number of new colormap entries we will allocate */
      theMaxCell = theColorSize[theRED] * theColorSize[theGREEN] *
                   theColorSize[theBLUE];
      _mWGetBestGraymap(theVisual);
      break;

    case StaticGray:
      theColormap = XCreateColormap(theDisplay,
                                    RootWindow(theDisplay, theScreen),
                                    theVisual, AllocNone);

      theMaxCell = theCells;
      theMaxDist = 0x7fffffff;
      if ((theColors = (XColor *)calloc(theCells, sizeof(XColor))) == NULL)
        _mWMAError("X local colormap", theCells * sizeof(XColor));
      for (i = 0; i < theCells; i++)
      {
        theColors[i].pixel = i;
      }
      if ((theColorCube = (XColor *)calloc(theGRAYSIZE, sizeof(XColor))) == NULL)
        _mWMAError("X gray ramp", theGRAYSIZE * sizeof(XColor));
      break;

    case TrueColor:
      theColormap = XCreateColormap(theDisplay,
                                    RootWindow(theDisplay, theScreen),
                                    theVisual, AllocNone);
      break;

    case DirectColor:
      theColormap = XCreateColormap(theDisplay,
                                    RootWindow(theDisplay, theScreen),
                                    theVisual, AllocNone);

      rmask = (theColorSize[theRED]   - 1)*(0x10000/theColorSize[theRED]);
      gmask = (theColorSize[theGREEN] - 1)*(0x10000/theColorSize[theGREEN]);
      bmask = (theColorSize[theBLUE]  - 1)*(0x10000/theColorSize[theBLUE]);

      /*
       * This should do the job of initializing a DirectColor visual, but
       * it has had relatively little testing, as I only have a single
       * 8-bit DirectColor visual to test it out on.  What's really lacking
       * is something smart to do if the XAllocColor() call fails.
       */
      for (i = 0; i < theCells; i++)
      {
        XColor tmp;
        tmp.red = tmp.green = tmp.blue = i * 0xffff / (theCells - 1);

        tmp.red   &= rmask;
        tmp.green &= gmask;
        tmp.blue  &= bmask;
        tmp.flags = DoRed | DoGreen | DoBlue;

        XAllocColor(theDisplay, theColormap, &tmp);
      }
      break;
  }
}



/*****************************************************************************
*
* FUNCTION GetBestColormap
*
* INPUT
*
*    theVisual - the type of visual to use
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
* DESCRIPTION
*
*   Builds a minimal palette for a PseudoColor visual to be filled as
*   needed as the rendering progresses.
*
* CHANGES
*
*   June, 1996: Changed to deferred colormap allocation [AED]
*
******************************************************************************/

static void _mWGetBestColormap(Visual *theVisual){
  unsigned long rd, gd, bd;

  /* We only need to allocate this once if we are doing an animation */
  if (theColors == NULL)
  {
    int index;
    if ((theColors = (XColor *)calloc(theCells, sizeof(XColor))) == NULL)
        _mWMAError("X local colormap", theCells * sizeof(XColor));
    for (index = 0; index < theCells; index++)
    {
      theColors[index].pixel = index;
    }

    if ((theColorCube = (XColor *)calloc(theCUBESIZE*theCUBESIZE*theCUBESIZE,
                                         sizeof(XColor))) == NULL)
      _mWMAError("X color cube",
              theCUBESIZE*theCUBESIZE*theCUBESIZE*sizeof(XColor));
  }

  if (!(theParameters & thePRIVATECMAP))
  {
    if ((theColormap = DefaultColormap(theDisplay, theScreen)) == None)
    {
      _mWError("\nUnable to get default X colormap.\n");
    }
  }
  else /* We aren't using the default visual, so we need a private colormap */
  {
    int low_colors;

    if (theColormap == None && (theColormap = XCreateColormap(theDisplay,
           RootWindow(theDisplay, theScreen), theVisual, AllocAll)) == None)
    {
      _mWError("\nUnable to get private X colormap.\n");
    }

    /*
     * We'll put the color cube in the top end of the colormap with the
     * intent of setting the lower cells to match those of the default
     * colormap.  The window manager and the clients that are already
     * running are hopefully using just these lower cells. After we copy
     * the colors in thse low cells from the default map to the new map,
     * there should be less color flashing.
     */

    theMaxCell = (theCells * 7 + 7) / 8;
    low_colors = theCells - theColorSize[theRED] * theColorSize[theGREEN] *
                                                   theColorSize[theBLUE];

    XQueryColors(theDisplay, DefaultColormap(theDisplay, theScreen),
                 theColors, low_colors);
    XStoreColors(theDisplay, theColormap, theColors, low_colors);
  }
  rd = 0x3000 / (theColorSize[theRED] - 1);
  gd = 0x3000 / (theColorSize[theGREEN] - 1);
  bd = 0x3000 / (theColorSize[theBLUE] - 1);

  theMaxDist = rd*rd + gd*gd + bd*bd;

  /* Make sure we get at least the eight corners of the color cube */
  for (rd = 0; rd < theCUBESIZE; rd += theCUBESIZE - 1)
  for (gd = 0; gd < theCUBESIZE; gd += theCUBESIZE - 1)
  for (bd = 0; bd < theCUBESIZE; bd += theCUBESIZE - 1)
    _mWGetBestIndex(rd ? 0xffff : 0, gd ? 0xffff : 0, bd ? 0xffff : 0,
                 &theColorCube[(rd * theCUBESIZE + gd) * theCUBESIZE + bd]);

  /* Go for middle gray as well */
  _mWGetBestIndex(0x8000, 0x8000, 0x8000,
               &theColorCube[((theCUBESIZE * theCUBESIZE + theCUBESIZE) *
                               theCUBESIZE + theCUBESIZE) / 2]);
}


/*****************************************************************************
*
* FUNCTION   GetBestGraymap
*
* INPUT      theVisual  -  visual to build gray map for
*
* OUTPUT
*
* RETURNS
*
* AUTHOR     Andreas Dilger
*
* DESCRIPTION
*
*   Builds a minimal grayscale palette for either a PseudoColor or
*   GrayScale visual that will be filled in as needed.
*
* CHANGES
*
*   June, 1996: Changed to deferred colormap allocation [AED]
*
******************************************************************************/

static void _mWGetBestGraymap(Visual *theVisual){
  if (theColors == NULL)
  {
    int index;
    if ((theColors = (XColor *)calloc(theCells, sizeof(XColor))) == NULL)
      _mWMAError("X local colormap", theCells * sizeof(XColor));
    for (index = 0; index < theCells; index++)
    {
      theColors[index].pixel = index;
    }

    if ((theColorCube = (XColor *)calloc(theGRAYSIZE, sizeof(XColor))) == NULL)
      _mWMAError("X gray ramp", theGRAYSIZE * sizeof(XColor));
  }

  /*
   * Get the default colormap, and try to use this first.
   */

  if (!(theParameters & thePRIVATECMAP))
  {
    if ((theColormap = DefaultColormap(theDisplay, theScreen)) == None)
    {
      _mWError("\nUnable to get default X colormap.\n");
    }
  }
  else /* We aren't using the default visual - create a private colormap */
  {
    int low_colors;

    if (theColormap == None && (theColormap = XCreateColormap(theDisplay,
           RootWindow(theDisplay, theScreen), theVisual, AllocAll)) == None)
    {
      _mWError("\nUnable to get private X colormap.\n");
    }

    /*
     * We'll put the gray ramp in the top end of the colormap with the
     * intent of setting the lower cells to match those of the default
     * colormap.  The window manager and the clients that are already
     * running are hopefully using just these lower cells. After we copy
     * the colors in thse low cells from the default map to the new map,
     * there should be less color flashing.
     */

    theMaxCell = (theCells * 7 + 7) / 8;
    low_colors = theCells - theColorSize[theRED] * theColorSize[theGREEN] *
                                                   theColorSize[theBLUE];

    XQueryColors(theDisplay, DefaultColormap(theDisplay, theScreen),
                 theColors, low_colors);
    /*
     * If we have a GrayScale visual, we may as well store the low colors
     * in their grayscale equivalent, as the server will only use the red
     * component anyways, and this will give us more colors to work with
     * without affecting the root window too much.
     */
    if (theClass == GrayScale)
    {
      unsigned long gray;
      int index;

      for (index = 0; index < low_colors; index++)
      {
        gray = (307 * theColors[index].red + 599 * theColors[index].green +
                118 * theColors[index].blue) / 1024;
        theColors[index].red =
        theColors[index].green =
        theColors[index].blue = gray;
      }
    }

    XStoreColors(theDisplay, theColormap, theColors, low_colors);
  }

  theMaxDist = 0x8000 / (theMaxCell - 1);
  theMaxDist = 3 * theMaxDist * theMaxDist;

  /* Make sure we get at least black, middle gray, and white */
  _mWGetBestIndex(0x0000, 0x0000, 0x0000, &theColorCube[0]);
  _mWGetBestIndex(0xffff, 0xffff, 0xffff, &theColorCube[theGRAYSIZE - 1]);
  _mWGetBestIndex(0x8000, 0x8000, 0x8000, &theColorCube[theGRAYSIZE / 2]);
}


/*****************************************************************************
*
* FUNCTION GetBestIndex
*
* INPUT r, g, b  - index into color cube to match
*
* OUTPUT
*
* RETURNS theCell - the ColorCube entry filled in with the best color match
*
* AUTHOR  Andreas Dilger.
*
* DESCRIPTION
*
*    Finds the nearest color in the given colormap using a euclidean distance.
*    This routine also does deferred colormap allocation based on requested
*    colors, rather than allocating a static color cube at the beginning of
*    the render.  If we can't find a colormap entry that is within the given
*    error bound, then we try to allocate the given color.  In any case, we
*    return the index of the color closest to that specified.
*
*    This is considerably faster than allocating a fixed set of colors at the
*    beginning and filling the color cube, as most images don't use the full
*    range of colors.  It also has the advantage that it is possible to get
*    more accurate colors in the colormap than a generic color cube.
*
* CHANGES
*
*    Changed for dynamic colormap allocation.  July 9, 1996 [AED]
*
******************************************************************************/

static void _mWGetBestIndex(long r, long g, long b, XColor *theCell){
  static int numAlloc = 0, count = 200;
  unsigned long bestDist = 0xffffffffL;
  int bestIndex = 0, newIndex = FALSE, nCell;

  /*
   * Get a copy of the colormap periodically.  I tried the ColormapNotify
   * Event, but it didn't seem to register a change in a colormap entry
   * like I thought it should.
   */
  if ((count++ > 20 && numAlloc%10 == 0)||(numAlloc < theCells && count > 200))
  {
    XQueryColors(theDisplay, theColormap, theColors, (int)theCells);

    /* The static visuals won't change, so pretend we allocated everything */
    if (theClass != PseudoColor && theClass != GrayScale)
    {
      numAlloc = theCells;
    }
    count = 0;
  }

  /*
   * Work out the Euclidean distance for each colormap entry.  We need
   * to divide these by 2, or the maximum possible distance won't fit
   * into a 32-bit unsigned long, and we don't really need floats for this.
   */

  for (nCell = 0; nCell < theCells; nCell++)
  {
    unsigned long dist;
    long rd, gd, bd;

    rd = (theColors[nCell].red   - r) / 2;
    gd = (theColors[nCell].green - g) / 2;
    bd = (theColors[nCell].blue  - b) / 2;
    dist = rd*rd + gd*gd + bd*bd;
    /* Is this the best match we've found so far? */
    if (dist < bestDist)
    {
      bestDist = dist;
      bestIndex = nCell;
    }
  }


  /*
   * If we couldn't find a close enough entry and we haven't already
   * allocated too many entries, or if we haven't allocated this entry
   * ourselves we will try to allocate the desired color now.
   */
  if ((numAlloc < theMaxCell && bestDist > theMaxDist) || bestIndex >= numAlloc)
  {
    XColor want;
    if (bestDist > theMaxDist)
    {
      want.red   = r;
      want.green = g;
      want.blue  = b;
      want.pixel = theCells - numAlloc - 1;
   }
    else
    {
      want = theColors[bestIndex];
    }
    want.flags = DoRed | DoGreen | DoBlue;

    /*
     * Either we are just storing the desired color in our private
     * colormap, or we are trying to allocate it in the shared colormap.
     */
    if (theColormap != DefaultColormap(theDisplay, theScreen))
    {
      XStoreColor(theDisplay, theColormap, &want);
      newIndex = TRUE;
    }
    else if (XAllocColor(theDisplay, theColormap, &want))
    {
      newIndex = TRUE;
    }

    /*
     * We need to update our local color table with the data for the
     * old location of this palette index.
     */
    if (newIndex)
    {
      for (nCell = numAlloc; nCell < theCells; nCell++)
      {
        if (want.pixel == theColors[nCell].pixel)
        {
          /* Save actual pixel colors in the next local color table entry */
          XColor tmp = theColors[numAlloc];

          theColors[numAlloc] = want;
          theColors[nCell] = tmp;
          bestIndex = numAlloc++;
          break;
        }
      }
    }
  }

  *theCell = theColors[bestIndex];
  /* This just tells use we have filled in this cell already, nothing else */
  theCell->flags = DoRed | DoGreen | DoBlue;
}



/*****************************************************************************
*
* FUNCTION GetBestColor
*
* INPUT r, g, b  - desired color to match in the range [0, 0xffff]
*
* OUTPUT
*
* RETURNS
*
*   pointer to the color which is closest, whether the output is grayscale
*   or color.
*
* AUTHOR
*
* DESCRIPTION
*
*    Returns the pointer from the color cube or ramp which was previously
*    calculated to be the closest to the desired color.  If we haven't
*    already found the closest color for this cell, then we need to find it
*    first.  One must be careful with the order of operations when working
*    out the index, to avoid overflow with large color and index values.
*
* CHANGES
*
*    Add deferred color cube/ramp allocation and filling here.
*
******************************************************************************/

static XColor *_mWGetBestColor (long r, long g, long b){
  unsigned long index;

    long rc, gc, bc;

    rc = r * theCUBESIZE / 0x10000;
    gc = g * theCUBESIZE / 0x10000;
    bc = b * theCUBESIZE / 0x10000;

    index = (rc * theCUBESIZE + gc) * theCUBESIZE + bc;

    /* Check if we have found the closest color yet */
    if (!theColorCube[index].flags)
    {
      _mWGetBestIndex(r, g, b, &theColorCube[index]);
    }

  return (&theColorCube[index]);
}


/*****************************************************************************
* Error
******************************************************************************/
int  _mWError(char *format,...){                                                                               
  va_list marker;                                                               
                                                                                
  va_start(marker, format);                                                     
  vfprintf(stderr, format, marker);                                           
  va_end(marker);                                                               
  exit(1);                                                              
}                                                                               

/*****************************************************************************
* Memory Allocation Error
******************************************************************************/
int _mWMAError(char *msg, int val) {
  fprintf(stderr,"Cannot allocate %d bytes for %s.\n",val,msg);
  exit(1);
}  

/*****************************************************************************
* Event Handler
******************************************************************************/
Bool MyWindows(Display *display, XEvent *event, XPointer arg) {
  mHandler *H;
  Bool result;
  
  if (!theHandlers) 
    return True;
  _LOCK
    for (H=theHandlers,result=False;H;H=H->next)
      if (H->W->theWindow==event->xany.window) {
        result=True;
	break;
      }
  _UNLOCK
  return result;
}

void * _mWHandleEvents(void *ptr){
  XEvent e;
  mHandler *H;

  while(1) {
    XIfEvent(theDisplay,&e,MyWindows,NULL);    
    if (!theHandlers) {handlerRunning=0;pthread_exit(NULL);}
    _LOCK
      for (H=theHandlers;H;H=H->next)
        if (H->W->theWindow==e.xany.window) break;
    _UNLOCK
    H->func(H->W,&e);
  }  
}
